#include "GameState.hpp"

// ...

#include "ShowActions.hpp"

namespace game_state
{
	EnvironmentG* gEnvironment = nullptr;
	Wizard* gWizard = nullptr;
	Player* gPlayer = nullptr;
	CurrentOpponent* gCurrOppo = nullptr;

	bool gFirst = true;
	uint32_t gTurn = 0;

	void CleanGlobals()
	{
		if (gEnvironment != nullptr)
		{
			delete gEnvironment;
			gEnvironment = nullptr;
		}

		if (gWizard != nullptr)
		{
			delete gWizard;
			gWizard = nullptr;
		}

		if (gPlayer != nullptr)
		{
			delete gPlayer;
			gPlayer = nullptr;
		}

		if (gCurrOppo != nullptr)
		{
			delete gCurrOppo;
			gCurrOppo = nullptr;
		}

		show::InitMessageQueue();

		gFirst = true;
		gTurn = 0;
		std::srand(1983); //TODO: should be 'time' after testing!
	}

	std::string GetName(const ShifterShape& shape)
	{
		switch (shape)
		{
		case ShifterShape::Cat: return "Cat";
		case ShifterShape::Wolf:  return "Wolf";
		case ShifterShape::Turtle:  return "Turtle";
		case ShifterShape::Bird:  return "Bird";
		case ShifterShape::Rat:  return "RAT";
		case ShifterShape::Minotaur:  return "MINOTAUR";
		case ShifterShape::Cultist:  return "CULTIST";
		case ShifterShape::Dragon:  return "DRAGON";
		case ShifterShape::Forrest:  return "[env]";
		case ShifterShape::Wall:  return "[env]";
		case ShifterShape::Dungeon:  return "[env]";
		case ShifterShape::Wizard:  return "WIZARD";
		default: return "BLOB?";
		}
	};


	std::pair<TurnResult, std::string> HandlePlayerChoice(const TurnChoice& player_choice)
	{
		EnvironmentG* p_env = EnvironmentG::GetInstance();
		Wizard* p_wizard = Wizard::GetInstance();
		Player* p_player = Player::GetInstance();
		CurrentOpponent* p_oppo = CurrentOpponent::GetInstance();

		std::string message("");

		//const auto mesg_env = p_env->PlayerEffect(player_choice);
		//if (!mesg_env.empty())
		//{
		//	show::PrintMessage("The lands " + mesg_env);
		//}
		const auto mesg_wiz = p_wizard->PlayerEffect(player_choice);
		if (!mesg_wiz.empty())
		{
			show::PrintMessage("The wizard " + mesg_wiz);
		}
		const auto mesg_plr = p_player->PlayerEffect(player_choice);
		if (!mesg_plr.empty())
		{
			show::PrintMessage("The familiar " + mesg_plr);
		}
		const auto mesg_oppo = p_oppo->PlayerEffect(player_choice);
		if (!mesg_oppo.empty())
		{
			show::PrintMessage(GetName(p_oppo->GetShape()) + " " + mesg_oppo);
		}

		return std::make_pair(p_oppo->FinalDeath() ? TurnResult::Win : (p_player->GetCurrentHitPoints() > 0 ? (p_oppo->GetCurrentHitPoints() > 0 ? TurnResult::Continue : TurnResult::Next) : TurnResult::Lose), message);
	}

	std::pair<TurnResult, std::string> HandleWizardChoice()
	{
		Wizard* p_wizard = Wizard::GetInstance();
		CurrentOpponent* p_oppo = CurrentOpponent::GetInstance();

		const StackAction wiz_action = p_wizard->GetActionStack()->Pop(gTurn);

		std::string message("");

		if (wiz_action == StackAction::Heal || wiz_action == StackAction::Defend)
		{
			const auto mesg_part = Player::GetInstance()->HitWith(wiz_action, { Modifier::Magical }, 30);
			if (!mesg_part.empty())
			{
				show::PrintMessage("The familiar " + mesg_part);
			}
		}
		else
		{
			const auto mesg_part_env = EnvironmentG::GetInstance()->HitWith(wiz_action, { Modifier::Magical }, 1);
			if (wiz_action != StackAction::Attack && ! mesg_part_env.empty())
			{
				show::PrintMessage("The environment " + mesg_part_env);
			}
			const auto mesg_part_oppo = p_oppo->HitWith(wiz_action, { Modifier::Magical }, 4);
			if (! mesg_part_oppo.empty())
			{
				show::ShowAttack(ShifterShape::Wizard, Modifier::Magical);
				show::PrintMessage(GetName(p_oppo->GetShape()) + " " + mesg_part_oppo);
			}
		}

		return std::make_pair(p_oppo->FinalDeath() ? TurnResult::Win : (p_oppo->GetCurrentHitPoints() > 0 ? TurnResult::Continue : TurnResult::Next), message);
	}

	std::pair<TurnResult, std::string> HandleOppoChoice()
	{
		EnvironmentG* p_env = EnvironmentG::GetInstance();
		Wizard* p_wizard = Wizard::GetInstance();
		Player* p_player = Player::GetInstance();
		CurrentOpponent* p_oppo = CurrentOpponent::GetInstance();

		const StackAction oppo_action = p_oppo->GetActionStack()->Pop(gTurn);

		std::string message("");

		if (oppo_action == StackAction::Heal || oppo_action == StackAction::Defend)
		{
			const auto mesg_part = p_oppo->HitWith(oppo_action, { Modifier::Magical }, 30);
			if (!mesg_part.empty())
			{
				show::PrintMessage(GetName(p_oppo->GetShape()) + " " + mesg_part);
			}
		}
		else
		{
			auto mods = p_env->GetModifiers();
			mods.insert(p_oppo->GetOffensiveModifier());

			const auto mesg_part_plr = p_player->HitWith(oppo_action, mods, p_oppo->GetCurrentAttack());
			if (!mesg_part_plr.empty())
			{
				show::ShowAttack(p_oppo->GetShape(), Modifier::Physical);
				show::PrintMessage("The familiar " + mesg_part_plr);
			}
			const auto mesg_part_env = p_env->HitWith(oppo_action, mods, 1);
			if (oppo_action != StackAction::Attack && ! mesg_part_env.empty())
			{
				show::PrintMessage("The surroundings " + mesg_part_env);
			}
		}

		return std::make_pair(p_player->GetCurrentHitPoints() > 0 ? (p_oppo->GetCurrentHitPoints() > 0 ? TurnResult::Continue : TurnResult::Next) : TurnResult::Lose, message);
	}

	std::pair<TurnResult, std::string> HandleTurnChange()
	{
		EnvironmentG* p_env = EnvironmentG::GetInstance();
		Wizard* p_wizard = Wizard::GetInstance();
		Player* p_player = Player::GetInstance();
		CurrentOpponent* p_oppo = CurrentOpponent::GetInstance();

		++gTurn;
		if (gFirst)
		{
			gTurn = 0;
			gFirst = false;
		}

		p_env->NextTurn();
		p_wizard->NextTurn();
		p_player->NextTurn();
		p_oppo->NextTurn();

		return std::make_pair(TurnResult::Continue, std::string(""));
	}

	/*static*/ StackAction BaseCreature::StayIdle(const uint32_t&)
	{
		return StackAction::Idle;
	}

	BaseCreature::BaseCreature(const behave_func_t& func_get_next) :
		next_actions_(func_get_next),
		hp_max_(1),
		hp_left_(1)
	{
		next_actions_.Fill(3);
	}

	BaseCreature::BaseCreature() :
		BaseCreature(BaseCreature::StayIdle)
	{
	}

	// NOTE: returns dead or not!
	/*virtual*/ bool BaseCreature::NextTurn()
	{
		atk_turn_ = atk_norm_;
		def_turn_ = def_norm_;
		weaknesses_turn_ = weaknesses_norm_;
		strengths_turn_ = strengths_norm_;

		return hp_left_ > 0; // TODO?: messages for returning to normal (like special conditions ... resist extra cold by spell e.d.)
	}

	// NOTE: returns message for player
	/*virtual*/ std::string BaseCreature::HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength)
	{
		std::string message;

		float env_mod = 1.f;
		for (const auto& mod : mods)
		{
			env_mod *= ApplyModifier(mod);
		}
		env_mod = std::fmin(4.f, env_mod);

		const int32_t hp_before = hp_left_;
		switch (action)
		{
		case StackAction::Attack:
			hp_left_ -= (int32_t)std::fmax(1.f, ((strength * (4.f - env_mod)) - (1.f * def_turn_)));
			message = "tries to withstand the attack";
			break;
		case StackAction::Defend:
			def_turn_ += strength;
			message = "prepares to defend";
			break;
		case StackAction::Dry:
			message = "shakes off the water";
			break;
		case StackAction::Wet:
			message = "sniffles";
			break;
		case StackAction::Cold:
			message = "shivers";
			break;
		case StackAction::Hot:
			message = "looks parched";
			break;
		case StackAction::Heal:
			hp_left_ += (int32_t)std::fmax(1.f, (1.f * strength));
			message = "'s wounds dissapear";
			break;
		case StackAction::EnvNormal:
			message = "checks the weather";
			break;
		case StackAction::EnvRandom:
			message = "frowns at the weather";
			break;
		case StackAction::ResistCold:
			weaknesses_turn_.erase(Modifier::Cold);
			strengths_turn_.insert(Modifier::Cold);
			message = "was never bothered by the cold anyway";
			break;
		case StackAction::ResistHot:
			weaknesses_turn_.erase(Modifier::Hot);
			strengths_turn_.insert(Modifier::Hot);
			message = "could walk over hot coals";
			break;
		case StackAction::ResistWet:
			weaknesses_turn_.erase(Modifier::Wet);
			strengths_turn_.insert(Modifier::Wet);
			message = "is ready for the rain";
			break;
		case StackAction::WeakToCold:
			weaknesses_turn_.insert(Modifier::Cold);
			strengths_turn_.erase(Modifier::Cold);
			message = "sneezes loudly";
			break;
		case StackAction::WeakToHot:
			weaknesses_turn_.insert(Modifier::Hot);
			strengths_turn_.erase(Modifier::Hot);
			message = "can't stand the heat";
			break;
		case StackAction::WeakToWet:
			weaknesses_turn_.insert(Modifier::Wet);
			strengths_turn_.erase(Modifier::Wet);
			message = "didn't remember to bring an umbrella";
			break;
		case StackAction::LessOffence:
			atk_turn_ -= (int32_t)std::fmax(1.f, 1.f * strength);
			if (atk_turn_ < 0)
			{
				atk_turn_ = 0;
			}
			message = "seemns uncertain";
			break;
		case StackAction::MoreOffence:
			atk_turn_ += (int32_t)std::fmax(1.f, 1.f * strength);
			message = "sports a nasty grin";
			break;
		case StackAction::LessDefence:
			def_turn_ -= (int32_t)std::fmax(1.f, 1.f * strength);
			if (def_turn_ < 0)
			{
				def_turn_ = 0;
			}
			message = "feels weaker somehow";
			break;
		case StackAction::MoreDefence:
			def_turn_ += (int32_t)std::fmax(1.f, 1.f * strength);
			message = "thoughens up";
			break;
		default:
			message = "[UNKOWN ACTION]";
			break;
		}

		const int32_t down_hp = hp_before - hp_left_;
		if (down_hp > 0)
		{
			if (env_mod < 1.f)
			{
				message += ", handles it easily";
			}
			else if (env_mod > 1.f)
			{
				message += ", doesn't like this sort of thing";
			}
		}

		if (hp_left_ > hp_max_)
		{
			hp_left_ = hp_max_;
		}

		return message;
	}

	// NOTE: > 1.f strength, < 1.f weakness
	/*virtual*/ float BaseCreature::ApplyModifier(const Modifier& env)
	{
		float mod = 1.f;

		if (weaknesses_turn_.count(env) > 0)
		{
			mod /= 2.f;
		}

		if (strengths_turn_.count(env) > 0)
		{
			mod *= 2.f;
		}

		return mod;
	}

	void BaseCreature::SetBaseStats(const int32_t& hp, const int32_t& atk, const int32_t& def)
	{
		hp_max_ = hp;
		atk_norm_ = atk;
		def_norm_ = def;
	}

	void BaseCreature::ResetStats()
	{
		hp_left_ = hp_max_;
		atk_turn_ = atk_norm_;
		def_turn_ = def_norm_;
	}

	float BaseCreature::GetPercentage()
	{
		return ((float) hp_left_) / hp_max_;
	}

	float CurrentOpponent::GetPercentageX()
	{
		return current_->GetPercentage();
	}

	int32_t BaseCreature::GetCurrentAttack() const
	{
		return atk_turn_;
	}

	int32_t BaseCreature::GetCurrentDefence() const
	{
		return def_turn_;
	}

	int32_t BaseCreature::GetCurrentHitPoints() const
	{
		return hp_left_;
	}

	/*virtual*/ Modifier BaseCreature::GetOffensiveModifier()
	{
		return Modifier::Physical;
	}

	/*virtual*/ std::string BaseCreature::PlayerEffect(const TurnChoice& player_choice)
	{
		return "";
	}

	/*virtual*/ ActionStack* BaseCreature::GetActionStack()
	{
		return &next_actions_;
	}

	ActionStack::ActionStack(const behave_func_t& func_get_next) :
		func_get_next_(func_get_next)
	{
	}

	void ActionStack::Fill(const uint32_t& num)
	{
		for (uint32_t i_turn; num > actions_.size(); ++i_turn)
		{
			actions_.push_back(func_get_next_(i_turn));
		}
	}

	StackAction ActionStack::Peek(const uint32_t& num) const
	{
		return (num >= actions_.size()) ? StackAction::Idle : actions_[num];
	}

	StackAction ActionStack::Pop(const uint32_t& turn)
	{
		StackAction result = StackAction::Idle;
		if (!actions_.empty())
		{
			result = actions_.back();
			actions_.pop_back();
		}
		actions_.push_front(func_get_next_(turn + actions_.size()));
		return result;
	}

	void ActionStack::Confuse()
	{
		if (actions_.size() < 3)
		{
			return;
		}
		//std::swap(actions_[1], actions_[2]);

		const auto temp_a = actions_.back();
		actions_.pop_front();
		const auto temp_b = actions_.back();
		actions_.pop_front();
		actions_.push_front(temp_a);
		actions_.push_front(temp_b);
	}

	void ActionStack::ReplaceLast(const StackAction& action)
	{
		if (actions_.empty())
		{
			return;
		}
		actions_.pop_front();
		actions_.push_front(action);
	}

	/*static*/ EnvironmentG* EnvironmentG::GetInstance()
	{
		if (gEnvironment == nullptr)
		{
			gEnvironment = new EnvironmentG();
		}
		return gEnvironment;
	}

	EnvironmentG::EnvironmentG() :
		BaseCreature(),
		looks_(ShifterShape::Forrest),
		humidity_(Modifier::Dry),
		temperature_(Modifier::Dry)
	{
	}

	// NOTE: returns dead or not! : env cant die, always true
	/*virtual*/ bool EnvironmentG::NextTurn()
	{
		return true;
	}

	// NOTE: returns message for player : env mods only!
	/*virtual*/ std::string EnvironmentG::HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength)
	{
		std::string message("");

		// NOTE: Don't call super, override everything.

		switch (action)
		{
		case StackAction::Dry:
			humidity_ = Modifier::Dry;
			message = "the rain stops";
			break;
		case StackAction::Wet:
			humidity_ = Modifier::Wet;
			message = "it starts raining harshly";
			break;
		case StackAction::Cold:
			temperature_ = Modifier::Cold;
			message = "a sudden cold snap begins";
			break;
		case StackAction::Hot:
			temperature_ = Modifier::Hot;
			message = "the sun seems intense";
			break;
		case StackAction::EnvNormal:
			humidity_ = Modifier::Dry;
			temperature_ = Modifier::Dry;
			message = "the surroundings return to normal";
			break;
		case StackAction::EnvRandom:
		{
			const int32_t rnd_hum = std::rand() % 2; //TODO: replace bad rand!
			const int32_t rnd_tmp = std::rand() % 3; //TODO: replace bad rand!
			humidity_ = (rnd_hum > 0) ? Modifier::Wet : Modifier::Dry;
			temperature_ = (rnd_tmp > 0) ? ((rnd_tmp > 1) ? Modifier::Cold : Modifier::Hot) : Modifier::Dry;
		}
		message = "the weather is shaken up";
		break;
		default:
			break;
		}

		return message;
	}

	const std::set<Modifier> EnvironmentG::GetModifiers() const
	{
		return{ humidity_, temperature_ };
	}

	/*virtual*/ ShifterShape EnvironmentG::GetShape()
	{
		return looks_;
	}

	/*virtual*/ bool EnvironmentG::TryShapeshift(const ShifterShape& shape)
	{
		switch (shape)
		{
		case ShifterShape::Forrest:  // intentional fallthrough!
		case ShifterShape::Wall:     // intentional fallthrough!
		case ShifterShape::Dungeon:
			looks_ = shape;
			return true;
		default:
			return false;
		}
	}

	/*static*/ Player* Player::GetInstance()
	{
		if (gPlayer == nullptr)
		{
			gPlayer = new Player();
			gPlayer->TryShapeshift(ShifterShape::Cat);
			gPlayer->ResetStats();
		}
		return gPlayer;
	}

	Player::Player() :
		BaseCreature() //,
		//skip_turns_(0)
	{
	}

	/*virtual*/ ShifterShape Player::GetShape()
	{
		return shape_;
	}

	bool Player::TryShapeshift(const ShifterShape& shape)
	{
		bool result = true;

		weaknesses_norm_.clear();
		strengths_norm_.clear();

		switch (shape)
		{
		case ShifterShape::Cat:
			SetBaseStats(100, 8, 10);
			weaknesses_norm_.insert(Modifier::Wet);
			break;
		case ShifterShape::Wolf:
			SetBaseStats(100, 10, 9);
			strengths_norm_.insert(Modifier::Cold);
			weaknesses_norm_.insert(Modifier::Hot);
			break;
		case ShifterShape::Turtle:
			SetBaseStats(100, 4, 15);
			weaknesses_norm_.insert(Modifier::Cold);
			break;
		case ShifterShape::Bird:
			SetBaseStats(100, 6, 12);
			weaknesses_norm_.insert(Modifier::Wet);
			break;
		default:
			result = false;
			break;
		}

		if (result)
		{
			shape_ = shape;
			show::ShowShift(shape_);
		}
		return result;
	}

	/*virtual*/ std::string Player::PlayerEffect(const TurnChoice& player_choice)
	{
		std::string message("");

		switch (player_choice)
		{
		case TurnChoice::Evade:
			def_turn_ += 2;
			break;
		case TurnChoice::Special:
			switch (shape_)
			{
			case ShifterShape::Cat:
				message = "shimmers magically";
				break;
			case ShifterShape::Wolf:
				atk_turn_ += 5;
				def_turn_ -= 2;
				message = "goes all out on the attack, choosing to forego defence";
				break;
			case ShifterShape::Turtle:
				def_turn_ += 4;
				message = "hides in that handy shell for a while";
				break;
			case ShifterShape::Bird:
				message = "flies by the opponent, unbalancing the opponent";
				break;
			}
			break;
		case TurnChoice::NormalAttack:
			message = "tries to attack";
			break;
		case TurnChoice::ShiftToCat:
			message = (shape_ == ShifterShape::Cat) ? "stays a cat" : "becomes a cat";
			TryShapeshift(ShifterShape::Cat);
			break;
		case TurnChoice::ShiftToWolf:
			message = (shape_ == ShifterShape::Wolf) ? "stays a wolf" : "becomes a wolf";
			TryShapeshift(ShifterShape::Wolf);
			break;
		case TurnChoice::ShiftToTurtle:
			message = (shape_ == ShifterShape::Turtle) ? "stays a turtle" : "becomes a turtle";
			TryShapeshift(ShifterShape::Turtle);
			break;
		case TurnChoice::ShiftToBird:
			message = (shape_ == ShifterShape::Bird) ? "stays a bird" : "becomes a bird";
			TryShapeshift(ShifterShape::Bird);
			break;
		case TurnChoice::Suggest:
			message = "asks the wizard for a boon";
			break;
		default:
			break;
		}

		if (def_turn_ < 0)
		{
			def_turn_ = 1;
		}
		if (atk_turn_ < 0)
		{
			atk_turn_ = 1;
		}

		return message;
	}

	std::string Player::Describe(const TurnChoice& player_choice) const
	{
		switch (player_choice)
		{
		case TurnChoice::NormalAttack:
			return "attack";
		case TurnChoice::Evade:
			return "evade";
		case TurnChoice::ShiftToCat:
			return "become a cat";
		case TurnChoice::ShiftToWolf:
			return "become a wolf";
		case TurnChoice::ShiftToTurtle:
			return "become a turtle";
		case TurnChoice::ShiftToBird:
			return "become a bird";
		case TurnChoice::Special:
			switch (shape_)
			{
			case ShifterShape::Cat:    return "feline magic (confuse opponent)";
			case ShifterShape::Wolf:   return "wolf all-out attack";
			case ShifterShape::Turtle: return "hide in shell this turn";
			case ShifterShape::Bird:   return "fly-by special attack (2 turns)";
			default: return "[NOT A PLAYER SHAPE]";
			}
		case TurnChoice::Suggest:
			switch (shape_)
			{
			case ShifterShape::Cat:    return "[cat] suggest harm opponent";
			case ShifterShape::Wolf:   return "[wolf] suggest cold weather";
			case ShifterShape::Turtle: return "[turtle] suggest healing spell";
			case ShifterShape::Bird:   return "[bird] suggest normal weather";
			default: return "[NOT A PLAYER SHAPE]";
			}
		default:
			return "[UNKOWN ACTION]";
		}
	}

	/*static*/ Wizard* Wizard::GetInstance()
	{
		if (gWizard == nullptr)
		{
			gWizard = new Wizard();
		}
		return gWizard;
	}

	Wizard::Wizard() :
		BaseCreature(Wizard::Behaviour)
	{
	}

	// NOTE: returns dead or not! : wizard cant die, in current design always true
	/*virtual*/ bool Wizard::NextTurn()
	{
		return true;
	}

	// NOTE: returns message for player : wizard doesn't really get hit in current design
	/*virtual*/ std::string Wizard::HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength)
	{
		return "";
	}

	/*virtual*/ std::string Wizard::PlayerEffect(const TurnChoice& player_choice)
	{
		std::string message("");

		if (player_choice == TurnChoice::Suggest)
		{
			Player* p_player = Player::GetInstance();

			switch (p_player->GetShape())
			{
			case ShifterShape::Cat:
				next_actions_.ReplaceLast(StackAction::Attack);
				message = "considers your request to magically harm the opponent";
				break;
			case ShifterShape::Wolf:
				next_actions_.ReplaceLast(StackAction::Cold);
				message = "thinks about making the weather more wolf-friendly";
				break;
			case ShifterShape::Turtle:
				next_actions_.ReplaceLast(StackAction::Heal);
				message = "assures you that your health is important";
				break;
			case ShifterShape::Bird:
				next_actions_.ReplaceLast(StackAction::EnvNormal);
				message = "is indeed annoyed by the weather, now that you mention it";
				break;
			}
		}

		return message;
	}

	/*virtual*/ ShifterShape Wizard::GetShape()
	{
		return ShifterShape::Wizard;
	}

	/*virtual*/ bool Wizard::TryShapeshift(const ShifterShape& shape)
	{
		return false;
	}

	/*virtual*/ Modifier Wizard::GetOffensiveModifier()
	{
		return Modifier::Magical;
	}

	/*static*/ StackAction Wizard::Behaviour(const uint32_t& turn)
	{
		Player* p_player = Player::GetInstance();

		switch (turn % 6)
		{
		case 0: return StackAction::Defend;
		case 1: return StackAction::Attack;
		case 2: return (p_player->GetCurrentHitPoints() < 15) ? StackAction::Heal : StackAction::Attack;
		case 3: return (p_player->GetShape() != ShifterShape::Wolf) ? StackAction::Hot : StackAction::Wet;
		case 4: return StackAction::Attack;
		case 5: return StackAction::EnvNormal;
		}
		return StackAction::Idle;
	}

	/*static*/ CurrentOpponent* CurrentOpponent::GetInstance()
	{
		if (gCurrOppo == nullptr)
		{
			gCurrOppo = new CurrentOpponent();
		}
		return gCurrOppo;
	}

	CurrentOpponent::CurrentOpponent() :
		BaseCreature(),
		current_(nullptr)
	{
		//current_ = new SingleOppo(ShifterShape::Rat, SingleOppo::RatBehaviourA);
		sorting_algo_of_evil_.push_back(new SingleOppo(ShifterShape::Rat, SingleOppo::RatBehaviourA));
		sorting_algo_of_evil_.push_back(new SingleOppo(ShifterShape::Minotaur, SingleOppo::MinotaurBehaviourA));
		sorting_algo_of_evil_.push_back(new SingleOppo(ShifterShape::Rat, SingleOppo::RatBehaviourA));
		sorting_algo_of_evil_.push_back(new SingleOppo(ShifterShape::Cultist, SingleOppo::CultistBehaviourA));
		sorting_algo_of_evil_.push_back(new SingleOppo(ShifterShape::Minotaur, SingleOppo::MinotaurBehaviourA));
		sorting_algo_of_evil_.push_back(new SingleOppo(ShifterShape::Cultist, SingleOppo::CultistBehaviourA));
		sorting_algo_of_evil_.push_back(new SingleOppo(ShifterShape::Dragon, SingleOppo::DragonBehaviourA));
		NextTurn();
	}

	/*virtual*/ std::string CurrentOpponent::HitWith(const StackAction& action, const std::set<Modifier>& mods, const int32_t& strength)
	{
		return current_->HitWith(action, mods, strength);
	}

	/*virtual*/ std::string CurrentOpponent::PlayerEffect(const TurnChoice& player_choice)
	{
		return current_->PlayerEffect(player_choice);
	}

	/*virtual*/ bool CurrentOpponent::NextTurn()
	{
		const bool result = current_ ? current_->NextTurn() : false;
		if ((! result) && (! FinalDeath()))
		{
			if (current_) // <-- == 'not first time'
			{
				delete current_;
			}

			current_ = sorting_algo_of_evil_.front();
			if (sorting_algo_of_evil_.size() > 1)
			{
				sorting_algo_of_evil_.pop_front();
			}
		}
		return result;
	}

	/*virtual*/ ShifterShape CurrentOpponent::GetShape()
	{
		return current_ ? current_->GetShape() : ShifterShape::Dragon;
	}

	/*virtual*/ bool CurrentOpponent::TryShapeshift(const ShifterShape& shape)
	{
		return false; // always?? --> FUTURE: Shifting opponents...
	}

	/*virtual*/ Modifier CurrentOpponent::GetOffensiveModifier()
	{
		return current_->GetOffensiveModifier();
	}

	bool CurrentOpponent::FinalDeath()
	{
		return (sorting_algo_of_evil_.size() <= 1) && (current_->GetCurrentHitPoints() <= 0) && (current_ == sorting_algo_of_evil_.front());
	}

	/*virtual*/ ActionStack* CurrentOpponent::GetActionStack()
	{
		return current_->GetActionStack();
	}

	/*virtual*/ int32_t CurrentOpponent::GetCurrentAttack() const
	{
		return current_->GetCurrentAttack();
	}

	/*virtual*/ int32_t CurrentOpponent::GetCurrentDefence() const
	{
		return current_->GetCurrentDefence();
	}

	/*virtual*/ int32_t CurrentOpponent::GetCurrentHitPoints() const
	{
		return current_->GetCurrentHitPoints();
	}

	SingleOppo::SingleOppo(const ShifterShape& looks, const behave_func_t& behave_func) :
		BaseCreature(behave_func)
	{
		const bool no_rat = looks != ShifterShape::Minotaur || looks != ShifterShape::Cultist || looks != ShifterShape::Dragon;
		looks_ = no_rat ? looks : ShifterShape::Rat;

		switch (looks_)
		{
		case ShifterShape::Rat:
			SetBaseStats(45, 3, 3);
			weaknesses_norm_.insert(Modifier::Wet);
			weaknesses_norm_.insert(Modifier::Hot);
			weaknesses_norm_.insert(Modifier::Cold);
			break;
		case ShifterShape::Minotaur:
			SetBaseStats(120, 6, 7);
			weaknesses_norm_.insert(Modifier::Cold);
			strengths_norm_.insert(Modifier::Hot);
			break;
		case ShifterShape::Cultist:
			SetBaseStats(70, 4, 8);
			weaknesses_norm_.insert(Modifier::Hot);
			strengths_norm_.insert(Modifier::Wet);
			strengths_norm_.insert(Modifier::Magical);
			break;
		case ShifterShape::Dragon:
			SetBaseStats(200, 8, 9);
			weaknesses_norm_.insert(Modifier::Magical);
			strengths_norm_.insert(Modifier::Hot);
			strengths_norm_.insert(Modifier::Physical);
			break;
		default: break;
		}

		ResetStats();
	}

	/*virtual*/ std::string SingleOppo::PlayerEffect(const TurnChoice& player_choice)
	{
		std::string message("");

		Player* p_player = Player::GetInstance();
		std::set<Modifier> mods = EnvironmentG::GetInstance()->GetModifiers();
		
		switch (player_choice)
		{
		case TurnChoice::NormalAttack:
			show::ShowAttack(p_player->GetShape(), Modifier::Physical);
			mods.insert(Modifier::Physical);
			message = this->HitWith(StackAction::Attack, mods, p_player->GetCurrentAttack());
			break;
		case TurnChoice::Special:
			switch (p_player->GetShape())
			{
			case ShifterShape::Cat:
				show::ShowAttack(ShifterShape::Cat, Modifier::Magical);
				next_actions_.Confuse();
				message = "messes up plans in confusion";
				break;
			case ShifterShape::Wolf:
				{
					show::ShowAttack(ShifterShape::Wolf, Modifier::Physical);
					show::ShowAttack(ShifterShape::Wolf, Modifier::Physical);
					auto mods = EnvironmentG::GetInstance()->GetModifiers();
					mods.insert(Modifier::Physical);
					HitWith(StackAction::Attack, mods, Player::GetInstance()->GetCurrentAttack());
				}
				message = "braces for impact";
				break;
			case ShifterShape::Turtle:
				message = "tries to think of a way to get you out of your shell";
				break;
			case ShifterShape::Bird:
				show::ShowAttack(ShifterShape::Bird, Modifier::Magical);
				def_turn_ -= 4;
				message = "swats anoyedly, but misses, at your repeated fly-by's";
				break;
			default:
				message = "[NOT A PLAYER SHAPE]";
				break;
			}
			break;
		default:
			break;
		}

		if (def_turn_ < 0)
		{
			def_turn_ = 1;
		}
		if (atk_turn_ < 0)
		{
			atk_turn_ = 1;
		}

		return message;
	}

	/*virtual*/ ShifterShape SingleOppo::GetShape()
	{
		return looks_;
	}

	/*virtual*/ bool SingleOppo::TryShapeshift(const ShifterShape& shape)
	{
		return false; // always?? --> FUTURE: Shifting opponents...
	}

	/*virtual*/ Modifier SingleOppo::GetOffensiveModifier()
	{
		switch (looks_)
		{
		case ShifterShape::Cultist:
			return Modifier::Magical;
		case ShifterShape::Dragon:
			return ((std::rand() % 2) == 0) ? Modifier::Magical : Modifier::Physical; // TODO: Get rid of bad rand!
		default:
			return Modifier::Physical;
		}
	}

	/*static*/ StackAction SingleOppo::RatBehaviourA(const uint32_t& turn)
	{
		switch (turn % 3)
		{
		case 0: return StackAction::Attack;
		case 1: return StackAction::Attack;
		case 2: return StackAction::Defend;
		}
		return StackAction::Idle;
	}

	/*static*/ StackAction SingleOppo::MinotaurBehaviourA(const uint32_t& turn)
	{
		switch (turn % 4)
		{
		case 0: return StackAction::Attack;
		case 1: return StackAction::Defend;
		case 2: return StackAction::Attack;
		case 3: return gCurrOppo != nullptr && CurrentOpponent::GetInstance()->GetCurrentHitPoints() < 10 ? StackAction::Defend : StackAction::Attack;
		}
		return StackAction::Idle;
	}

	/*static*/ StackAction SingleOppo::CultistBehaviourA(const uint32_t& turn)
	{
		switch (turn % 5)
		{
		case 0: return StackAction::EnvRandom;
		case 1: return StackAction::Attack;
		case 2: return StackAction::Heal;
		case 3: return gCurrOppo != nullptr && CurrentOpponent::GetInstance()->GetCurrentHitPoints() < 15 ? StackAction::Defend : StackAction::Attack;
		case 4: return gCurrOppo != nullptr && CurrentOpponent::GetInstance()->GetCurrentHitPoints() < 15 ? StackAction::Defend : StackAction::Attack;
		}
		return StackAction::Idle;
	}

	/*static*/ StackAction SingleOppo::DragonBehaviourA(const uint32_t& turn)
	{
		switch (turn % 6)
		{
		case 0: return StackAction::Hot;
		case 1: return StackAction::Attack;
		case 2: return StackAction::Defend;
		case 3: return gCurrOppo != nullptr && CurrentOpponent::GetInstance()->GetCurrentHitPoints() < 100 ? StackAction::Defend : StackAction::Attack;
		case 4: return gCurrOppo != nullptr && CurrentOpponent::GetInstance()->GetCurrentHitPoints() <  25 ? StackAction::Defend : StackAction::Attack;
		case 5: return StackAction::EnvNormal;
		}
		return StackAction::Idle;
	}


}
