#include "ShowActions.hpp"

#include "GameState.hpp"
#include "main.hpp"

namespace show
{
	using namespace game_state;

	class MessageQueue
	{
	public:
		MessageQueue() :
			show_max_(5)
		{
		}

		void Clear()
		{
			//for (auto img : msgs_)
			//{
			//	delete img;
			//}
			msgs_.clear();
		}

		void PushMessage(const std::string& message, const SDL_Color& col = { 0xD0, 0xD0, 0xD0 })
		{
			MyImage img;
			if (!MySdlMedia::GetInstance()->CreateFromText(message, col, &img))
			{
				std::fprintf(stderr, "Error: Could not create mesg.queue text: %s\n", message.c_str());
				return;
			}
			while (msgs_.size() > show_max_)
			{
				// delete mesg_.front();
				msgs_.pop_front();
			}
			msgs_.push_back(img);
		}

		void Show(const int32_t& w, const int32_t& h)
		{
			auto screen = MySdlCore::GetInstance();

			int32_t y = h;
			for (auto& img : msgs_)
			{
				screen->DrawImage({ w, y, img.Width(), img.Height() }, img);
				y += 20;
			}
		}

	private:
		int32_t show_max_;
		std::deque<MyImage> msgs_;
	} gMessageQueue;

	class TextMenu
	{
	public:
		TextMenu(const std::vector<std::string>& opts) :
			selected_(0)
		{
			longest_width_ = 0;
			for (const std::string& opt : opts)
			{
				MyImage img;
				if (!MySdlMedia::GetInstance()->CreateFromText("    " + opt, { 0xCC, 0xCC, 0xCC }, &img))
				{
					std::fprintf(stderr, "Error: Could not create menu text: %s\n", opt.c_str());
					continue;
				}

				MyImage sec;
				if (!MySdlMedia::GetInstance()->CreateFromText("<*> " + opt, { 0xFF, 0xAA, 0xFF }, &sec))
				{
					std::fprintf(stderr, "Error: Could not create menu text: %s\n", opt.c_str());
					continue;
				}

				opts_.push_back(img);
				opts_sec_.push_back(sec);

				if (sec.Width() > longest_width_)
				{
					longest_width_ = sec.Width();
				}
			}
			max_ = (int32_t)opts_.size();
		}

		TextMenu() :
			selected_(-1),
			max_(-1),
			opts_({}),
			opts_sec_({}),
			longest_width_(0)
		{
		}

		void Down()
		{
			selected_ = (selected_ + 1) % max_;
		}

		void Up()
		{
			selected_ = (selected_ + (max_ - 1)) % max_;
		}

		void Left()
		{
			selected_ = (selected_ + (max_ - (max_ / 2))) % max_;
		}

		void Right()
		{
			selected_ = (selected_ + (max_ / 2)) % max_;
		}

		void View(const int& x, const int& y)
		{
			auto screen = MySdlCore::GetInstance();

			const int32_t hmax = max_ / 2;
			for (int32_t i_row = 0; i_row < hmax; ++i_row)
			{
				const int32_t i_row2 = i_row + hmax;

				MyImage& img = (i_row == selected_) ? opts_sec_[i_row] : opts_[i_row];
				screen->DrawImage({ x, y + (i_row * 20), img.Width(), img.Height() }, img);

				MyImage& img2 = (i_row2 == selected_) ? opts_sec_[i_row2] : opts_[i_row2];
				screen->DrawImage({ x + longest_width_ + 2, y + (i_row * 20), img2.Width(), img2.Height() }, img2);
			}
		}

		int32_t Selection()
		{
			return selected_;
		}

	private:
		int32_t selected_;
		int32_t max_;

		std::vector<MyImage> opts_;
		std::vector<MyImage> opts_sec_;

		int32_t longest_width_;
	};

	TextMenu gMenus[2];

	void InitMainMenu()
	{
		gMenus[0] = TextMenu
			(
		{
			"[1] help     | ",
			"[2] new game | ",
			"[3] continue   ",
			"[q] quit       "
		}
		);
	}

	void InitGameMenu()
	{
		gMenus[1] = TextMenu
			(
		{
			"[1] ATTACK      | ",
			"[2] EVADE       | ",
			"[3] SPECIAL     | ",
			"[4] SUGGEST     | ",
			"[5] TURN CAT      ",
			"[6] TURN WOLF     ",
			"[7] TURN TURTLE   ",
			"[8] TURN BIRD     "
		}
		);
	}

	void InitMessageQueue()
	{
		gMessageQueue.Clear();
	}

	void MainMenuDown()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[0].Down();
	}

	void MainMenuUp()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[0].Up();
	}

	void MainMenuLeft()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[0].Left();
	}

	void MainMenuRight()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[0].Right();
	}

	int32_t GetMainMenuChoice()
	{
		return gMenus[0].Selection() + 1;
	}

	void ShowMainMenu(const int32_t& w, const int32_t& h)
	{
		gMenus[0].View(w, h);
	}

	void GameMenuUp()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[1].Up();
	}

	void GameMenuDown()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[1].Down();
	}

	void GameMenuLeft()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[1].Left();
	}

	void GameMenuRight()
	{
		MySdlCore::GetInstance()->PlaySound(MySdlMedia::GetInstance()->GetSound(SndId::FxSelect));
		gMenus[1].Right();
	}

	int32_t GetGameMenuChoice()
	{
		return gMenus[1].Selection() + 1;
	}

	void ShowGameMenu(const int32_t& w, const int32_t& h)
	{
		gMenus[1].View(w, h);
	}

	void PrintMessage(const std::string& message)
	{
		gMessageQueue.PushMessage(message);
	}

	void ShowMessages(const int32_t& w, const int32_t& h)
	{
		gMessageQueue.Show(w, h);
	}

	void ShowAttack(const ShifterShape& shape, const Modifier& mod)
	{
		auto screen = MySdlCore::GetInstance();
		auto media = MySdlMedia::GetInstance();

		SDL_Rect from = { 600, 500, 120, 120 };
		SDL_Rect to = { 200, 200, 20, 20 };
		SDL_Rect from2 = { 200, 500, 120, 120 };
		SDL_Rect to2 = { 600, 200, 20, 20 };

		switch (shape)
		{
		case ShifterShape::Cat:
			screen->PlaySound(media->GetSound(mod == Modifier::Magical ? SndId::FxMagic : SndId::FxCat));
			screen->AnimateImage(from, to, media->GetImage(ImgId::CatAtk));
			break;
		case ShifterShape::Wolf:
			screen->PlaySound(media->GetSound(SndId::FxWolf));
			screen->AnimateImage(from, to, media->GetImage(ImgId::WolfAtk));
			break;
		case ShifterShape::Turtle:
			screen->PlaySound(media->GetSound(SndId::FxTurtle));
			screen->AnimateImage(from, to, media->GetImage(ImgId::TurtleAtk));
			break;
		case ShifterShape::Bird:
			screen->PlaySound(media->GetSound(SndId::FxBird));
			screen->AnimateImage(from, to, media->GetImage(ImgId::BirdAtkA));
			screen->AnimateImage(from2, to2, media->GetImage(ImgId::BirdAtkB));
			break;
		case ShifterShape::Rat:
			screen->PlaySound(media->GetSound(SndId::FxOppo));
			screen->AnimateImage(from, to, media->GetImage(ImgId::ClawAtk));
			break;
		case ShifterShape::Minotaur:
			screen->PlaySound(media->GetSound(SndId::FxOppo));
			screen->AnimateImage(from, to, media->GetImage(ImgId::HammerAtk));
			break;
		case ShifterShape::Cultist:
			screen->PlaySound(media->GetSound(SndId::FxOppo));
			screen->AnimateImage(from, to, media->GetImage(mod == Modifier::Magical ? ImgId::ShowBoom : ImgId::ClawAtk));
			break;
		case ShifterShape::Dragon:
			screen->PlaySound(media->GetSound(SndId::FxOppo));
			screen->AnimateImage(from, to, media->GetImage(mod == Modifier::Magical ? ImgId::ShowBoom : ImgId::ClawAtk));
			break;
		default:
			screen->PlaySound(media->GetSound(SndId::FxMagic));
			screen->AnimateImage(from, to, media->GetImage(mod == Modifier::Magical ? ImgId::ShowBoom : ImgId::ClawAtk));
			break;
		}
		//TODO: Add anims. for opponent as well.

		//SDL_Delay(96);
	}

	void ShowShift(const game_state::ShifterShape& shape)
	{
		auto screen = MySdlCore::GetInstance();
		auto media = MySdlMedia::GetInstance();

		screen->PlaySound(media->GetSound(SndId::FxMagic));

		//SDL_Delay(96);

		switch (shape)
		{
		case ShifterShape::Cat:
			screen->PlaySound(media->GetSound(SndId::MuzCat));
			// TODO!!!: Animations!
			break;
		case ShifterShape::Wolf:
			screen->PlaySound(media->GetSound(SndId::MuzWolf));
			break;
		case ShifterShape::Turtle:
			screen->PlaySound(media->GetSound(SndId::MuzTurtle));
			break;
		case ShifterShape::Bird:
			screen->PlaySound(media->GetSound(SndId::MuzBird));
			break;
		default:
			break;
		}
	}

	void WaitForKey()
	{
		auto screen = MySdlCore::GetInstance();
		auto media = MySdlMedia::GetInstance();
		screen->DrawImage({42, 213, 390, 111}, media->GetImage(ImgId::PressKey), true);

		SDL_Event evt;
		bool go = true;
		while (go)
		{
			while (SDL_PollEvent(&evt))
			{
				if (evt.type == SDL_QUIT)
				{
					std::exit(0);
				}
				else if (evt.type == SDL_KEYDOWN)
				{
					if (evt.key.keysym.sym == SDLK_q)
					{
						std::exit(0);
					}

					screen->PlaySound(media->GetSound(SndId::FxSelect));
					go = false;
				}
				else
				{
					SDL_Delay(32);
				}
			}
		}
	}

} //namespace show
