#include "main.hpp"

#include "GameState.hpp"
#include "ShowActions.hpp"

#include <cstdio>

ScreenMode gPrevScreenMode = ScreenMode::Lose;
ScreenMode gScreenMode = ScreenMode::MainMenu;

bool gGameFirstStep = true;
MyImage gGameCurrentBg; // = MySdlMedia::GetInstance()->GetImage(GetRandBg(ImgId::MainMenu));
bool gKeyRegistered = false;

TTF_Font *gFont = nullptr;

void ParkSdl()
{
	MySdlMedia::GetInstance()->Park();
	MySdlCore::GetInstance()->Park();
	SDL_Quit();
}

namespace my_sdl_internal
{
	MySdlCore* gCoreSingle = nullptr;
	MySdlMedia* gMediaSingle = nullptr;
}

	MyImage::MyImage() :
		tex_(nullptr),
		width_(-1),
		height_(-1)
	{
	}

	//MyImage::~MyImage()
	//{
	//	Free();
	//}

	bool MyImage::Set(SDL_Surface* surface)
	{
		tex_ = MySdlCore::GetInstance()->CreateTexture(surface);
		if (! tex_)
		{
			std::fprintf(stderr, "SDL unable to create texture: %s\n", SDL_GetError());
			return false;
		}
		width_ = surface->w;
		height_ = surface->h;
		return true;
	}

	SDL_Texture* MyImage::GetTexture() const
	{
		return tex_;
	}

	void MyImage::Free()
	{
		if (tex_)
		{
			SDL_DestroyTexture(tex_);
			tex_ = nullptr;
		}
	}

	int32_t MyImage::Width() const
	{
		return width_;
	}

	int32_t MyImage::Height() const
	{
		return height_;
	}

MySound::MySound() :
    is_sfx_(true),
    sfx_(nullptr),
	muz_(nullptr)
{
}

//MySound::~MySound()
//{
//	Free();
//}

bool MySound::Set(Mix_Chunk* sound)
{
	Free();
	is_sfx_ = true;
	sfx_ = sound;
	return true;
}

bool MySound::Set(Mix_Music* sound)
{
	Free();
	is_sfx_ = false;
	muz_ = sound;
	return true;
}

void MySound::Free()
{
	if (sfx_)
	{
        Mix_FreeChunk(sfx_);
		sfx_ = nullptr;
	}
	if (muz_)
	{
		Mix_FreeMusic(muz_);
		muz_ = nullptr;
	}
}

bool MySound::IsSfx() const
{
	return is_sfx_;
}

int32_t MySound::Play() const
{
	if (is_sfx_)
	{
		Mix_PlayChannel(-1, sfx_, 0);
		return 0;
	}
	else
	{
		const int32_t err_code = Mix_PlayMusic(muz_, -1);
		return err_code < 0 ? err_code : 1;
	}
}

bool MySound::Stop() const
{
	if (Mix_PlayingMusic() && ! is_sfx_)
	{
		Mix_HaltMusic();
	}
	return true;
}

	/*static*/ MySdlCore* MySdlCore::GetInstance()
	{
		using namespace my_sdl_internal;
		if (gCoreSingle == nullptr)
		{
			gCoreSingle = new MySdlCore();
			if (! gCoreSingle->BootSystem())
			{
				delete gCoreSingle;
				gCoreSingle = nullptr;
			}
		}
		return gCoreSingle;
	}

	SDL_PixelFormat* MySdlCore::GetPixelFormat() const
	{
		return bgSurface_->format;
	}

	bool MySdlCore::BootSystem()
	{
		// -- OVERALL --

		if (SDL_Init(SDL_INIT_EVERYTHING) != 0)
		{
			std::fprintf(stderr, "SDL init failed: %s\n", SDL_GetError());
			return false;
		}

		// -- WINDOW & VISUALS --

		mainWindow_ = SDL_CreateWindow("Familiar's Shape -- LD#35 -- rburema", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, winWidth_, winHeight_, SDL_WINDOW_SHOWN);
		if (! mainWindow_)
		{
			std::fprintf(stderr, "SDL create window failed: %s\n", SDL_GetError());
			Park();
			return false;
		}

		// - extended image loading -

		int32_t imgFlags = IMG_INIT_PNG;
		if (!(IMG_Init(imgFlags) & imgFlags))
		{
			std::fprintf(stderr, "SDL_image failed to init extended image support: %s\n", IMG_GetError());
			Park();
			return false;
		}
		img_inited_ = true;

		// - fonts loading -

		if (TTF_Init() != 0)
		{
			std::fprintf(stderr, "SDL_ttf failed to init fonts-subsystem: %s\n", TTF_GetError());
			Park();
			return false;
		}

		gFont = TTF_OpenFont(resc::gFontRef.c_str(), 18);
		if (! gFont)
		{
			std::fprintf(stderr, "SDL_ttf failed to load font-resource: %s\n", TTF_GetError());
			Park();
			return false;
		}

		ttf_inited_ = true;

		mainRenderer_ = SDL_CreateRenderer(mainWindow_, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
		if (! mainRenderer_)
		{
			std::fprintf(stderr, "SDL create renderer failed: %s\n", SDL_GetError());
			Park();
			return false;
		}

		bgSurface_ = SDL_GetWindowSurface(mainWindow_);

		// -- AUDIO --

		if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, MIX_DEFAULT_CHANNELS, 2048) != 0)
		{
			std::fprintf(stderr, "SDL_mixer failed to init audio-subsystem: %s\n", Mix_GetError());
			//Park();
			//return false;
		}
		else
		{
			can_play_audio_ = true;
		}

		return true;
	}

	void MySdlCore::ClearBackground() const
	{
		//SDL_FillRect(bgSurface_, nullptr, SDL_MapRGB(bgSurface_->format, 0xFF, 0xFF, 0xFF));
		//SDL_UpdateWindowSurface(mainWindow_);
		SDL_SetRenderDrawColor(mainRenderer_, 0xFF, 0xFF, 0xFF, 0x80);
		SDL_RenderClear(mainRenderer_);
		SDL_RenderPresent(mainRenderer_);
	}

	void MySdlCore::SetBackground(const MyImage& image, const bool& last_layer)
	{
		auto tex = image.GetTexture();
		if (tex)
		{
			SDL_RenderCopy(mainRenderer_, tex, NULL, NULL);
			if (last_layer)
			{
				SDL_RenderPresent(mainRenderer_);
			}
		}
		else
		{
			std::fprintf(stderr, "Could not display backgorund!");
		}
	}

	void MySdlCore::DrawImage(const SDL_Rect& rect, const MyImage& image, const bool& last_layer)
	{
		auto tex = image.GetTexture();
		if (tex)
		{
			SDL_RenderCopy(mainRenderer_, tex, NULL, &rect);
			if (last_layer)
			{
				SDL_RenderPresent(mainRenderer_);
			}
		}
		else
		{
			std::fprintf(stderr, "Could not display backgorund!");
		}
	}

	void MySdlCore::AnimateImage(const SDL_Rect& from, const SDL_Rect& to, const MyImage& image, const int32_t& steps, const int32_t& delay_msec)
	{
		auto tex = image.GetTexture();
		if (! tex)
		{
			return;
		}

		SDL_Surface* win_srf = SDL_GetWindowSurface(mainWindow_);

		SDL_Surface* temp = SDL_CreateRGBSurface(0, winWidth_, winHeight_, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000);
		SDL_RenderReadPixels(mainRenderer_, NULL, SDL_PIXELFORMAT_ARGB8888, temp->pixels, temp->pitch);
		SDL_Texture* ttemp = CreateTexture(temp);
		SDL_FreeSurface(temp);

		SDL_Rect incr = { (to.x - from.x) / steps, (to.y - from.y) / steps, (to.w - from.w) / steps, (to.h - from.h) / steps };

		SDL_Rect goal_rect(from);
		for (int32_t i_step = 0; i_step < steps; ++i_step)
		{
			SDL_RenderCopy(mainRenderer_, tex, NULL, &goal_rect);
			SDL_RenderPresent(mainRenderer_);
			SDL_Delay(delay_msec);
			SDL_RenderCopy(mainRenderer_, ttemp, NULL, NULL);

			goal_rect.x += incr.x;
			goal_rect.y += incr.y;
			goal_rect.w += incr.w;
			goal_rect.h += incr.h;
		}
		SDL_RenderPresent(mainRenderer_);

		SDL_DestroyTexture(ttemp);
	}

	void MySdlCore::FullRender()
	{
		SDL_RenderPresent(mainRenderer_);
	}

	void MySdlCore::PlaySound(const MySound& sound)
	{
		if (! can_play_audio_)
		{
			return;
		}

		if ((! sound.IsSfx()) && Mix_PlayingMusic())
		{
			sound.Stop(); // hackish
		}
		sound.Play();
	}

	void MySdlCore::Park()
	{
		if (can_play_audio_) { Mix_CloseAudio(); }
		if (ttf_inited_) { if (gFont) { TTF_CloseFont(gFont); gFont = nullptr; } TTF_Quit(); }
		if (img_inited_) { IMG_Quit(); }
		if (mainRenderer_) { SDL_DestroyRenderer(mainRenderer_); }
		if (mainWindow_) { SDL_DestroyWindow(mainWindow_); }
	}

	SDL_Texture* MySdlCore::CreateTexture(SDL_Surface* surface)
	{
		return SDL_CreateTextureFromSurface(mainRenderer_, surface);
	}

	MySdlCore::MySdlCore() :
		mainWindow_(nullptr),
		mainRenderer_(nullptr),
		winWidth_(960),
		winHeight_(720),
		can_play_audio_(false),
		current_muz_(nullptr),
		img_inited_(false),
		ttf_inited_(false)
	{
	}

	/*static*/ MySdlMedia* MySdlMedia::GetInstance()
	{
		using namespace my_sdl_internal;
		if (gMediaSingle == nullptr)
		{
			gMediaSingle = new MySdlMedia();

			for (const auto& img_loc : resc::gImageRefs)
			{
				if (! gMediaSingle->PutImage(img_loc.first, img_loc.second))
				{
					std::fprintf(stderr, "Couldn't load image-resource from path %s\n", img_loc.second.c_str());
				}
			}

			for (const auto& sfx_loc : resc::gSfxRefs)
			{
				if (! gMediaSingle->PutSound(true, sfx_loc.first, sfx_loc.second))
				{
					std::fprintf(stderr, "Couldn't load sfx-resource from path %s\n", sfx_loc.second.c_str());
				}
			}

			for (const auto& muz_loc : resc::gMusicRefs)
			{
				if (! gMediaSingle->PutSound(false, muz_loc.first, muz_loc.second))
				{
					std::fprintf(stderr, "Couldn't load music-resource from path %s\n", muz_loc.second.c_str());
				}
			}
		}
		return gMediaSingle;
	}

	bool MySdlMedia::PutImage(const ImgId& ref_name, const std::string& path)
	{
		SDL_Surface* loadSurface = IMG_Load(path.c_str());
		if (! loadSurface)
		{
			std::fprintf(stderr, "SDL_image unable to load image %s because: %s\n", path.c_str(), IMG_GetError());
			return false;
		}

		const MyImage empty_image;
		images_.insert(std::make_pair(ref_name, empty_image));
		MyImage& image = images_[ref_name];

		const bool success = image.Set(loadSurface);
		SDL_FreeSurface(loadSurface);
		return success;
	}

	bool MySdlMedia::CreateFromText(const std::string& message, const SDL_Color& color, MyImage* result)
	{
		SDL_Surface* textSurface = TTF_RenderText_Solid(gFont, message.c_str(), color);
		if (! textSurface)
		{
			fprintf(stderr, "SDL_ttf unable to render text to img: %s\n", TTF_GetError());
			return false;
		}
		
		result[0] = MyImage();
		const bool success = result[0].Set(textSurface);
		SDL_FreeSurface(textSurface);
		return success;
	}

	bool MySdlMedia::PutSound(const bool& is_sfx, const SndId& ref_name, const std::string& path)
	{
		bool success = true;

		const MySound empty_sound;
		sounds_.insert(std::make_pair(ref_name, empty_sound));
		MySound& sound = sounds_[ref_name];

		if (is_sfx)
		{
			Mix_Chunk* chunk = Mix_LoadWAV(path.c_str());
			success = (chunk != nullptr);
			sound.Set(chunk);
		}
		else
		{
			Mix_Music* muz = Mix_LoadMUS(path.c_str());
			success = (muz != nullptr);
			sound.Set(muz);
		}

		return success;
	}

	MyImage MySdlMedia::GetImage(const ImgId& ref_name)
	{
		if (images_.count(ref_name) < 1)
		{
			std::fprintf(stderr, "Missing ImgId %d\n", ref_name);
			MyImage dummy;
			return dummy;
		}
		return images_[ref_name];
	}

	MySound MySdlMedia::GetSound(const SndId& ref_name)
	{
		if (sounds_.count(ref_name) < 1)
		{
			std::fprintf(stderr, "Missing SndId %d\n", ref_name);
			MySound dummy;
			return dummy;
		}
		return sounds_[ref_name];
	}

	void MySdlMedia::Park()
	{
		for (auto& image : images_)
		{
			image.second.Free();
		}
		images_.clear();

		for (auto& sound : sounds_)
		{
			sound.second.Free();
		}
		sounds_.clear();
	}


	MySdlMedia::MySdlMedia()
    {
	}

// --- --- --- --- ---

ImgId GetRandBg()
{
	static std::vector<ImgId> backgrounds = { ImgId::EnvA, ImgId::EnvB, ImgId::EnvC };
	static ImgId last = ImgId::EnvC;

	auto choice = backgrounds[std::rand() % backgrounds.size()];  //TODO: replace bad rand!
	if (last == choice)
	{
		 choice = GetRandBg();
	}

	last = choice;
	return choice;
}

void DisplayMuzzle()
{
	auto media = MySdlMedia::GetInstance();
	auto screen = MySdlCore::GetInstance();

	using namespace game_state;
	switch (Player::GetInstance()->GetShape())
	{
	case ShifterShape::Cat: screen->SetBackground(media->GetImage(ImgId::ViewCat)); break;
	case ShifterShape::Wolf: screen->SetBackground(media->GetImage(ImgId::ViewWolf)); break;
	case ShifterShape::Turtle: screen->SetBackground(media->GetImage(ImgId::ViewTurtle)); break;
	case ShifterShape::Bird: screen->SetBackground(media->GetImage(ImgId::ViewBird)); break;
	}
}

void DisplayOppo()
{
	auto media = MySdlMedia::GetInstance();
	auto screen = MySdlCore::GetInstance();

	using namespace game_state;
	switch (CurrentOpponent::GetInstance()->GetShape())
	{
	case ShifterShape::Rat: screen->SetBackground(media->GetImage(ImgId::OppoRat)); break;
	case ShifterShape::Minotaur: screen->SetBackground(media->GetImage(ImgId::OppoMino)); break;
	case ShifterShape::Cultist: screen->SetBackground(media->GetImage(ImgId::OppoCult)); break;
	case ShifterShape::Dragon: screen->SetBackground(media->GetImage(ImgId::OppoDrag)); break;
	}
}

ImgId ActionToImg(game_state::StackAction action)
{
	switch (action)
	{
	case game_state::StackAction::Hot: return ImgId::IcoHot;
	case game_state::StackAction::Cold: return ImgId::IcoCold;
	case game_state::StackAction::Dry: return ImgId::IcoDry;
	case game_state::StackAction::Wet: return ImgId::IcoWet;
	case game_state::StackAction::Heal: return ImgId::IcoHeal;
	case game_state::StackAction::EnvNormal: return ImgId::IcoNorm;
	case game_state::StackAction::EnvRandom: return ImgId::IcoRnd;
	case game_state::StackAction::Attack: return ImgId::IcoAtk;
	case game_state::StackAction::Defend: return ImgId::IcoDef;
	default: return ImgId::BarGreen;
	}
}

void DisplayInGameScreen(const uint64_t& frame)
{
    static bool first = true;
	static MyImage current_mg = MySdlMedia::GetInstance()->GetImage(ImgId::EnvWiz);
	static MyImage current_vg = MySdlMedia::GetInstance()->GetImage(ImgId::EnvMenu);
	if (first)
	{
		first = false;
		gGameCurrentBg  = MySdlMedia::GetInstance()->GetImage(ImgId::EnvA);
	}

	auto media = MySdlMedia::GetInstance();
	auto screen = MySdlCore::GetInstance();

	screen->SetBackground(gGameCurrentBg);
	DisplayOppo();
	screen->SetBackground(current_mg);
	DisplayMuzzle();
	screen->SetBackground(current_vg);
	show::ShowMessages(40,  70);
	show::ShowGameMenu(50, 230);
	
	screen->DrawImage({ 20,  16, (int32_t) (game_state::CurrentOpponent::GetInstance()->GetPercentageX() * 920.f), 25}, media->GetImage(ImgId::BarRed));
	screen->DrawImage({ 20, 684, (int32_t) (game_state::Player::GetInstance()->GetPercentage() * 920.f), 25 }, media->GetImage(ImgId::BarGreen));

	auto wiz_stack = game_state::Wizard::GetInstance()->GetActionStack();
	screen->DrawImage({ 32, 420, 64, 64 }, media->GetImage(ActionToImg(wiz_stack->Peek(0))));
	screen->DrawImage({ 32, 510, 64, 64 }, media->GetImage(ActionToImg(wiz_stack->Peek(1))));
	screen->DrawImage({ 32, 600, 64, 64 }, media->GetImage(ActionToImg(wiz_stack->Peek(2))));

	auto oppo_stack = game_state::CurrentOpponent::GetInstance()->GetActionStack();
	screen->DrawImage({ 860, 64, 64, 64 }, media->GetImage(ActionToImg(oppo_stack->Peek(0))));
	screen->DrawImage({ 860, 154, 64, 64 }, media->GetImage(ActionToImg(oppo_stack->Peek(1))));
	screen->DrawImage({ 860, 244, 64, 64 }, media->GetImage(ActionToImg(oppo_stack->Peek(2))));


	// TODO? (here?) : Starts Animation & c...
}

bool CheckTurnPartResult(const std::pair<game_state::TurnResult, std::string>& turn_res, const bool& wait_key)
{
	using namespace game_state;

	static bool lock_bg = false;

	//show::PrintMessage(turn_res.second); //std::printf(turn_res.second.c_str());

	switch (turn_res.first)
	{
	case TurnResult::Next:
		if (!lock_bg)
		{
			game_state::EnvironmentG::GetInstance()->HitWith(StackAction::EnvNormal, {});
			show::InitMessageQueue();
			gGameCurrentBg = MySdlMedia::GetInstance()->GetImage(GetRandBg());
			lock_bg = true;
		}
		break; //return true;
	case TurnResult::Continue:
		lock_bg = false;
		break; //return true;
	case TurnResult::Win:
		lock_bg = false;
		gScreenMode = ScreenMode::Win;
		break; //return true; //return false;
	case TurnResult::Lose:
		lock_bg = false;
		gScreenMode = ScreenMode::Lose;
		break; //return true; //return false;
	default:
		lock_bg = false;
		std::fprintf(stderr, "\n[ERROR]\n");
		break; //return true;
	}

	if (wait_key)
	{
		DisplayInGameScreen(0);
		show::WaitForKey();
	}
	return true;
}

bool GameStep()
{
	if (gScreenMode != gPrevScreenMode)
	{
		gPrevScreenMode = gScreenMode;
		DisplayInGameScreen(0);
	}

	static uint64_t frame = 0;
	
	if (gGameFirstStep)
	{
		// ??? --> reinit??
		gGameFirstStep = false;
	}

	// -- -- --

	bool result = true;

	SDL_Event evt;
	while (SDL_PollEvent(&evt))
	{
		switch (evt.type)
		{
		case SDL_QUIT:
			return false;
			break;
		case SDL_KEYUP:
			gKeyRegistered = false;
			break;
		case SDL_KEYDOWN:
			if (! gKeyRegistered)
		    {
				gKeyRegistered = true;

				using namespace game_state;

				TurnChoice choice(TurnChoice::OoC);

				std::printf("OPPONENT: %d\n", CurrentOpponent::GetInstance()->GetCurrentHitPoints()); //TODO: Oppo name
				//std::printf("Wizard: \n");
				std::printf("Familiar: %d\n", Player::GetInstance()->GetCurrentHitPoints());
				//std::printf("Weather: \n");

				SDL_Keycode kcode = evt.key.keysym.sym;

				switch (kcode)
				{
				case SDLK_RETURN:
				case SDLK_SPACE:
					switch (show::GetGameMenuChoice())
				    {
					case 1: choice = TurnChoice::NormalAttack; break;
					case 2: choice = TurnChoice::Evade; break;
					case 3: choice = TurnChoice::Special; break;
					case 4: choice = TurnChoice::Suggest; break;
					case 5: choice = TurnChoice::ShiftToCat; break;
					case 6: choice = TurnChoice::ShiftToWolf; break;
					case 7: choice = TurnChoice::ShiftToTurtle; break;
					case 8: choice = TurnChoice::ShiftToBird; break;
					}
					break;
				case SDLK_w: case SDLK_UP: show::GameMenuUp(); break;
				case SDLK_s: case SDLK_DOWN: show::GameMenuDown(); break;
				case SDLK_a: case SDLK_LEFT: show::GameMenuLeft(); break;
				case SDLK_d: case SDLK_RIGHT: show::GameMenuRight(); break;
				}

				switch (kcode)
				{
				case SDLK_1: choice = TurnChoice::NormalAttack; break;
				case SDLK_2: choice = TurnChoice::Evade; break;
				case SDLK_3: choice = TurnChoice::Special; break;
				case SDLK_4: choice = TurnChoice::Suggest; break;
				case SDLK_5: choice = TurnChoice::ShiftToCat; break;
				case SDLK_6: choice = TurnChoice::ShiftToWolf; break;
				case SDLK_7: choice = TurnChoice::ShiftToTurtle; break;
				case SDLK_8: choice = TurnChoice::ShiftToBird; break;
				//case SDLK_9: choice = TurnChoice::NormalAttack; break;
				//case SDLK_0: choice = TurnChoice::NormalAttack; break;
				case SDLK_ESCAPE: gScreenMode = ScreenMode::MainMenu; break;
				case SDLK_q: result = false; break;
				}

				if (choice != TurnChoice::OoC)
				{
					result =
						CheckTurnPartResult(HandlePlayerChoice(choice), true) &&
						CheckTurnPartResult(HandleWizardChoice(), true) &&
						CheckTurnPartResult(HandleOppoChoice(), true) &&
						CheckTurnPartResult(HandleTurnChange(), false);
				}
		    }
			break;
		}
	}

	DisplayInGameScreen(frame);

	MySdlCore::GetInstance()->FullRender();
	SDL_Delay(32);
	++frame;
	return result;
}

void DrawMainMenu(const bool& res_muz)
{
	auto media = MySdlMedia::GetInstance();
	auto screen = MySdlCore::GetInstance();
	screen->SetBackground(media->GetImage(ImgId::MainMenu));
	show::ShowMainMenu(30, 30);
	if (res_muz)
	{
		screen->PlaySound(media->GetSound(SndId::MuzMenu));
	}
}

bool MenuStep()
{
	if (gScreenMode != gPrevScreenMode)
	{
		DrawMainMenu(true);
		gPrevScreenMode = gScreenMode;
	}

	bool result = true;

	SDL_Event evt;
	while (SDL_PollEvent(&evt))
	{
		switch (evt.type)
		{
		case SDL_QUIT:
			return false;
			break;
		case SDL_KEYUP:
			gKeyRegistered = false;
			switch (evt.key.keysym.sym)
			{
			case SDLK_w:
			case SDLK_UP:
				show::MainMenuUp();
				DrawMainMenu(false);
				break;
			case SDLK_s:
			case SDLK_DOWN:
				show::MainMenuDown();
				DrawMainMenu(false);
				break;
			case SDLK_d:
			case SDLK_LEFT:
				show::MainMenuLeft();
				DrawMainMenu(false);
				break;
			case SDLK_a:
			case SDLK_RIGHT:
				show::MainMenuRight();
				DrawMainMenu(false);
				break;
			case SDLK_RETURN:
			case SDLK_SPACE:
			    {
				    const int32_t viz_choice = show::GetMainMenuChoice();
					switch (viz_choice)
					{
					case 1:
						std::fprintf(stderr, "TODO: HELP SCREENS!\n");
						break;
					case 2:
						game_state::CleanGlobals();
						// NOTE: fallthrough!
					case 3:
						gScreenMode = ScreenMode::InGame;
						break;
					case 4:
						result = false;
						break;
					}
				    //DrawMainMenu();
			    }
				break;
			case SDLK_1:
				std::fprintf(stderr, "TODO: HELP SCREENS!\n");
				break;
			case SDLK_2:
				game_state::CleanGlobals();
				// NOTE: fallthrough!
			case SDLK_3:
				gScreenMode = ScreenMode::InGame;
				break;
			case SDLK_q:
				result = false;
				break;
			}
			break;
		case SDL_KEYDOWN:
			if (!gKeyRegistered)
			{
				gKeyRegistered = true;
			}
		}
	}

	MySdlCore::GetInstance()->FullRender();
	SDL_Delay(32);
	return result;
}

bool HelpStep()
{
	//TODO!

	// 	MySdlCore::GetInstance()->FullRender();

	return GameStep();
}

bool LoseStep()
{
	if (gScreenMode != gPrevScreenMode)
	{
		game_state::CleanGlobals();
		gGameFirstStep = true;

		auto media = MySdlMedia::GetInstance();
		auto screen = MySdlCore::GetInstance();
		screen->SetBackground(media->GetImage(ImgId::LoseScreen));
		screen->PlaySound(media->GetSound(SndId::MuzTurtle));
		screen->PlaySound(media->GetSound(SndId::FxOppo));
		gPrevScreenMode = gScreenMode;
	}

	bool result = true;

	SDL_Event evt;
	while (SDL_PollEvent(&evt))
	{
		switch (evt.type)
		{
		case SDL_QUIT:
			return false;
			break;
		case SDL_KEYUP:
			gKeyRegistered = false;
			switch (evt.key.keysym.sym)
			{
			case SDLK_ESCAPE:
			//case SDLK_SPACE:
				gScreenMode = ScreenMode::MainMenu;
				break;
			case SDLK_q:
				result = false;
				break;
			}
			break;
		case SDL_KEYDOWN:
			if (!gKeyRegistered)
			{
				gKeyRegistered = true;
			}
		}
	}

	MySdlCore::GetInstance()->FullRender();
	SDL_Delay(32);
	return result;
}

bool WinStep()
{
	if (gScreenMode != gPrevScreenMode)
	{
		game_state::CleanGlobals();
		gGameFirstStep = true;

		auto media = MySdlMedia::GetInstance();
		auto screen = MySdlCore::GetInstance();
		screen->SetBackground(media->GetImage(ImgId::WinScreen));
		screen->PlaySound(media->GetSound(SndId::FxMagic));
		screen->PlaySound(media->GetSound(SndId::MuzBird));
		gPrevScreenMode = gScreenMode;
	}

	bool result = true;

	SDL_Event evt;
	while (SDL_PollEvent(&evt))
	{
		switch (evt.type)
		{
		case SDL_QUIT:
			return false;
			break;
		case SDL_KEYUP:
			gKeyRegistered = false;
			switch (evt.key.keysym.sym)
			{
			case SDLK_ESCAPE:
				gScreenMode = ScreenMode::MainMenu;
				break;
			case SDLK_q:
				result = false;
				break;
			}
			break;
		case SDL_KEYDOWN:
			if (!gKeyRegistered)
			{
				gKeyRegistered = true;
			}
		}
	}

	MySdlCore::GetInstance()->FullRender();
	SDL_Delay(32);
	return result;
}

bool Step()
{
	switch (gScreenMode)
	{
	case ScreenMode::MainMenu: return MenuStep();
	case ScreenMode::Help: return HelpStep();
	case ScreenMode::InGame: return GameStep();
	case ScreenMode::Lose: return LoseStep();
	case ScreenMode::Win: return WinStep();
	default: std::fprintf(stderr, "ScreenMode not known!\n"); return false;
	}
}

//MyAni::MyAni(const ImgId& image, const std::vector<MyKeyFrame>& frames, const bool& loops) :
//    start_frame_(0),
//    image_(&MySdlMedia::GetInstance()->GetImage(image)),
//	loops_(loops)
//{
//	last_frame_ = 0;
//	for (auto& keyFrame : frames)
//	{
//		frames_.insert(std::make_pair(keyFrame.start, keyFrame));
//		last_frame_ = last_frame_ < keyFrame.start ? keyFrame.start : last_frame_;
//	}
//}
//
//void MyAni::Start(const uint64_t& current_frame)
//{
//	start_frame_ = current_frame;
//}
//
//bool MyAni::Display(const uint64_t& current_frame)
//{
//	uint32_t frame_diff = (uint32_t) (current_frame - start_frame_);
//	if (loops_)
//	{
//		frame_diff = frame_diff % (last_frame_ + 1);
//	}
//	auto& it_frame = frames_.lower_bound(frame_diff);
//	if (it_frame != frames_.end())
//	{
//		const MyKeyFrame& keyFrame = it_frame->second;
//
//	}
//	return loops_ || (frame_diff <= last_frame_);
//}

int main(int argc, char** argv)
{
	if (! MySdlCore::GetInstance())
	{
		std::fprintf(stderr, "Quit on boot error.");
		return 1;
	}

	if (! MySdlMedia::GetInstance())
	{
		std::fprintf(stderr, "Quit on load-media error.");
		return 1;
	}

	game_state::CleanGlobals();

	show::InitMainMenu();
	show::InitGameMenu();

	while (Step());

	ParkSdl(); // Not really nesc.
	return 0;
}
