#ifndef HDR_MAIN_HPP__
#define HDR_MAIN_HPP__

#include <cstdint>
#include <map>
#include <SDL.h>
#include <SDL_image.h>
#include <SDL_mixer.h>
#include <SDL_ttf.h>
#include <string>
#include <utility>
#include <vector>

enum class ImgId
{
	EnvA, EnvB, EnvC,
	EnvWiz, EnvMenu,
	MiscVfx,
	OppoRat, OppoMino, OppoCult, OppoDrag,
	ViewCat, ViewWolf, ViewTurtle, ViewBird,
	MainMenu,
	HelpA, HelpB, HelpC,
	LoseScreen, WinScreen,
	CatAtk, WolfAtk, TurtleAtk, BirdAtkA, BirdAtkB,
	ClawAtk, HammerAtk,
	ShowBoom,
	PressKey,
	BarRed, BarGreen,
	//
	IcoHot,
	IcoCold,
	IcoAtk, 
	IcoDef, 
	IcoNorm,
	IcoWet, 
	IcoDry, 
	IcoRnd, 
	IcoHeal,
	//
	HelpScreen,
};

enum class SndId
{
	FxSelect,
	FxMagic,
	FxCat, FxWolf, FxTurtle, FxBird,
	FxOppo,

	MuzMenu,
	MuzTense,
	MuzCat, MuzWolf, MuzTurtle, MuzBird
};

enum class ScreenMode
{
	MainMenu,
	Help,
	InGame,
	Lose,
	Win
};

namespace resc
{
	const std::map<ImgId, std::string> gImageRefs =
	{
		std::make_pair(ImgId::EnvA         , "./resc/img/bg_a_0.png"),
		std::make_pair(ImgId::EnvB         , "./resc/img/bg_b_0.png"),
		std::make_pair(ImgId::EnvC         , "./resc/img/bg_c_1.png"),
		std::make_pair(ImgId::EnvWiz       , "./resc/img/bg_x_1_wizard.png"),
		std::make_pair(ImgId::EnvMenu      , "./resc/img/bg_x_2_overlay.png"),
		std::make_pair(ImgId::MiscVfx      , "./resc/img/misc_vfx.png"),  // << -- << -- TOOD: REPLACE !!
		std::make_pair(ImgId::OppoRat      , "./resc/img/oppo_b_rat.png"),
		std::make_pair(ImgId::OppoMino     , "./resc/img/oppo_a_minotaur.png"),
		std::make_pair(ImgId::OppoCult     , "./resc/img/oppo_c_cultist.png"),
		std::make_pair(ImgId::OppoDrag     , "./resc/img/oppo_d_dragon.png"),
		std::make_pair(ImgId::ViewCat      , "./resc/img/view_cat_0.png"),
		std::make_pair(ImgId::ViewWolf     , "./resc/img/view_wolf_0.png"),
		std::make_pair(ImgId::ViewTurtle   , "./resc/img/view_turtle_0.png"),
		std::make_pair(ImgId::ViewBird     , "./resc/img/view_bird_0.png"),
		std::make_pair(ImgId::MainMenu     , "./resc/img/bg_mmenu.png"),
		std::make_pair(ImgId::LoseScreen   , "./resc/img/bg_end_lost.png"),
		std::make_pair(ImgId::WinScreen    , "./resc/img/bg_end_win.png"),
		std::make_pair(ImgId::HelpA        , "./resc/img/misc_vfx.png"),  // << -- TODO
		std::make_pair(ImgId::HelpB        , "./resc/img/misc_vfx.png"),  // << -- TODO
		std::make_pair(ImgId::HelpC        , "./resc/img/misc_vfx.png"),  // << -- TODO
		std::make_pair(ImgId::CatAtk       , "./resc/img/atk_cat.png"),
		std::make_pair(ImgId::WolfAtk      , "./resc/img/atk_wolf.png"),
		std::make_pair(ImgId::TurtleAtk    , "./resc/img/atk_turtle.png"),
		std::make_pair(ImgId::BirdAtkA     , "./resc/img/atk_bird_r.png"),
		std::make_pair(ImgId::BirdAtkB     , "./resc/img/atk_bird_l.png"),
		std::make_pair(ImgId::ClawAtk      , "./resc/img/atk_claw.png"),
		std::make_pair(ImgId::HammerAtk    , "./resc/img/atk_hammer.png"),
		std::make_pair(ImgId::ShowBoom     , "./resc/img/atk_bang.png"),
		std::make_pair(ImgId::PressKey     , "./resc/img/press_key.png"),
		std::make_pair(ImgId::BarRed       , "./resc/img/red.png"),
		std::make_pair(ImgId::BarGreen     , "./resc/img/green.png"),

		std::make_pair(ImgId::IcoHot, "./resc/img/ico_hot.png"),
		std::make_pair(ImgId::IcoCold, "./resc/img/ico_hot.png"),
		std::make_pair(ImgId::IcoAtk, "./resc/img/ico_atk.png"),
		std::make_pair(ImgId::IcoDef, "./resc/img/ico_def.png"),
		std::make_pair(ImgId::IcoNorm, "./resc/img/ico_wnorm.png"),
		std::make_pair(ImgId::IcoWet, "./resc/img/ico_wet.png"),
		std::make_pair(ImgId::IcoDry, "./resc/img/ico_dry.png"),
		std::make_pair(ImgId::IcoRnd, "./resc/img/ico_rnd.png"),
		std::make_pair(ImgId::IcoHeal, "./resc/img/ico_heal.png"),

		std::make_pair(ImgId::HelpScreen, "./resc/img/help_screen.png"),

	};

	const std::map<SndId, std::string> gSfxRefs =
	{
		std::make_pair(SndId::FxSelect  , "./resc/sfx/fx_select.wav"),
		std::make_pair(SndId::FxMagic   , "./resc/sfx/fx_magic.wav"),
		std::make_pair(SndId::FxCat     , "./resc/sfx/fx_cat_attack.wav"),
		std::make_pair(SndId::FxWolf    , "./resc/sfx/fx_wolf_attack.wav"),
		std::make_pair(SndId::FxTurtle  , "./resc/sfx/fx_turtle_attack.wav"),
		std::make_pair(SndId::FxBird    , "./resc/sfx/fx_bird_attack.wav"),
		std::make_pair(SndId::FxOppo    , "./resc/sfx/fx_oppo_attack.wav"),
	};

	const std::map<SndId, std::string> gMusicRefs =
	{
		std::make_pair(SndId::MuzMenu   , "./resc/muz/menu_music.wav"),
		std::make_pair(SndId::MuzTense  , "./resc/muz/tense_music.wav"),
		std::make_pair(SndId::MuzCat    , "./resc/muz/cat_song.wav"),
		std::make_pair(SndId::MuzWolf   , "./resc/muz/wolf_song.wav"),
		std::make_pair(SndId::MuzTurtle , "./resc/muz/turtle_song.wav"),
		std::make_pair(SndId::MuzBird   , "./resc/muz/bird_song.wav"),
	};

	const std::string gFontRef = "./resc/font/FantasqueSansMono-Regular.ttf";
}

class MyImage
{
public:
	MyImage();
	//~MyImage();

	bool Set(SDL_Surface* surface_);
	SDL_Texture* GetTexture() const;
	void Free();
	int32_t Width() const;
	int32_t Height() const;

private:
	int32_t width_;
	int32_t height_;
	SDL_Texture* tex_;
};

class MySound
{
public:
	MySound();
	//~MySound();

	bool Set(Mix_Chunk* sound);
	bool Set(Mix_Music* sound);
	void Free();
	bool IsSfx() const;
	int32_t Play() const;
	bool Stop() const;

private:
	bool is_sfx_;
	Mix_Chunk* sfx_;
	Mix_Music* muz_;
};

class MySdlCore
{
public:
	static MySdlCore* GetInstance();
	SDL_PixelFormat* GetPixelFormat() const;
	bool BootSystem();
	void ClearBackground() const;
	void SetBackground(const MyImage& image, const bool& last_layer = false);
	void DrawImage(const SDL_Rect& rect, const MyImage& image, const bool& last_layer = false);
	void AnimateImage(const SDL_Rect& from, const SDL_Rect& to, const MyImage& image, const int32_t& steps = 8, const int32_t& delay_msec = 16);
	void FullRender();
	void PlaySound(const MySound& sound);
	void Park();

	SDL_Texture* CreateTexture(SDL_Surface* surface);

private:
	MySdlCore();

	SDL_Window* mainWindow_;
	SDL_Renderer* mainRenderer_;
	SDL_Surface* bgSurface_;

	uint32_t winWidth_;
	uint32_t winHeight_;

	bool can_play_audio_;
	MySound* current_muz_;

	bool img_inited_;
	bool ttf_inited_;
};

class MySdlMedia
{
public:
	static MySdlMedia* GetInstance();
	bool PutImage(const ImgId& ref_name, const std::string& path);
	bool PutSound(const bool& is_sfx, const SndId& ref_name, const std::string& path);
	MyImage GetImage(const ImgId& ref_name);
	MySound GetSound(const SndId& ref_name);
	bool CreateFromText(const std::string& message, const SDL_Color& color, MyImage* result);
	void Park();
	
private:
	MySdlMedia();

	std::map<ImgId, MyImage> images_;
	std::map<SndId, MySound> sounds_;
};

//struct MyKeyFrame
//{
//public:
//	uint32_t start;
//	SDL_Rect place;
//};
//
//class MyAni
//{
//public:
//	MyAni(const ImgId& image, const std::vector<MyKeyFrame>& frames, const bool& loops);
//
//	void Start(const uint64_t& current_frame);
//	bool Display(const uint64_t& current_frame);
//
//private:
//	uint64_t start_frame_; // updated when plays
//	uint32_t last_frame_; // constant(-ish?)
//	MyImage* image_;
//	bool loops_;
//	std::map<uint32_t, MyKeyFrame> frames_;
//};

#endif // HDR_MAIN_HPP__
